(***********************************************************************)
(*                                                                     *)
(*                             Pio (OCaml)                             *)
(*                                                                     *)
(*                    Polyglot Input/Output Library                    *)
(*                                                                     *)
(*                  projet Estime, INRIA-Rocquencourt                  *)
(*                                                                     *)
(*  Copyright 2004-2011 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: io.mli,v 1.4 2011-09-13 11:41:23 fclement Exp $ *)

(** {3 The [Pio] library specification} *)

(**
   The [PIO] system is a set of communication libraries to safely exchange
   values between programs written in various languages.

   For the time being [PIO] offers a library for Caml, C, C++, and Fortran.
*)

(** {6 The [PIO] general API} *)

(**
   We describe here the overall definition and structure of the Application
   Programmer Interface, or API, of the [PIO] library, Version 1.0.
*)

(** {7 The concept of communication} *)

(**
   The [PIO] library defines and implements values, called
   ``Communications'', that can be safely exchanged between programs written
   in different languages.

   Any communication that satisfies the following specification (both the
   syntactic specification and the semantic restrictions) must be accepted
   without errors (except for hardware problems on the wires).
   The semantics of the transmitted values must not be modified by the
   transmission process: ping-pong communications should be a no-op (i.e. the
   value read back is equal to the value sent).

   We exchange communications as described below. Communication basic blocks
   are named {e typed values}; a typed value is a value tagged with its type.
*)

(** {7 The description of communications} *)

(**
   A communication can be:
   - a [Phrase] communication, i.e. a sequence of typed values,
   - a [Task] communication, i.e. a function name and its arguments,
   - a [Result] communication, i.e. a sequence of typed values,
   - an [Error] communication, i.e. an exception value,
   - a [Service] communication, i.e. a meta information about the
     communication.

   A [Task] communication is supposed to specify a function to be applied to
   its arguments by the receiver; the function is given by its name (a
   string) and the arguments (a sequence of typed values).

   The intended meaning of [Result] and [Error] communications is to tag the
   effect of a previous [Task] communication: if the function application
   succeeds, the result is sent back as a [Result] communication, otherwise
   an exception value is sent back using an [Error] communication.

   A [Service] communication is supposed to carry some information about the
   communication process: acknowledgement of communication, request for
   beginning/end of communication, request for initialization/finalization of
   the reader program, and the library error communication about bogus
   communication; each service message may have some associated typed values.

   If any data in the communication is not compliant to the grammar and to the
   various constraints expressed in the comments, then the entire communication
   is invalid. In that case, the library should not produce any value (even
   possibly truncated ones) but must return [Service "WrongCommunication"] to
   the calling program.

   The library communicates binary values encoded in so-called ``little''
   endian. Both ends of the communication have not to worry about endianness:
   the library takes care of the endianness when appropriate.

   The Caml implementation of [Communication] uses buffers. The user must
   provide appropriate i/o buffers (see the Buffer and Scanf.Scanning modules).
*)

(** {6 The grammar of communications} *)

(** {7 Communication description} *)

(**
   We describe here the precise string of characters that defines a valid
   communication.

   Communication ::=
     '(' '\n'                          (* Beginning of communication marker *)
     Phrase | Task | Result | Error | Service
     ')' '\n' '\n'                     (* End of communication marker *)

   Phrase ::=
     "%p" Values_number n '\n'         (* Phrase tag and number of values [n] *)
     Typed_value*                      (* [n] typed values *)

   Task ::=
     "%t" Values_number n ' ' Task_name '\n'
                                       (* Task tag, number of arguments [n], and
                                          task, or function, name *)
     Typed_value*                      (* [n] typed values *)

   Result ::=
     "%r" Values_number n '\n'         (* Result tag and number of results [n] *)
     Typed_value*                      (* [n] typed values *)

   Error ::=
     "%e" Values_number n '\n'         (* Error tag and number of values [n] *)
     Typed_value*                      (* [n] typed values *)

   Service ::=
     "%s" Values_number n ' ' Service_name '\n'
                                       (* Service tag, number of arguments [n],
                                          and service name *)
     Typed_value*                      (* [n] typed values *)

   Task_name ::= String                (* The name of the task is a string. *)

   Values_number ::= ' ' Size ' '      (* A [values_number] is
                                          a [Size] surrounded with spaces. *)

   Size ::= '<' Natural '>'            (* A size is a delimited natural. *)

   Natural ::= Decimal_digit+          (* A natural number in the usual 10
                                          basis representation. *)

   Decimal_digit ::= '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'
*)

(** {7 Service description} *)

(**
   Service_name ::=                    (* Intended meaning from the client's
                                          (the program that reads the
                                          communication) point of view. *)
   | "Ok"                              (* Acknowledge service: the
                                          preceding communication was
                                          understood. *)
   | "Ko"                              (* Negative-acknowledge service: the
                                          preceding communication was not
                                          understood. *)
   | "Allo"                            (* Communication requested. *)
   | "Bye"                             (* End of communication requested. *)
   | "Start"                           (* Initialization of program requested. *)
   | "Stop"                            (* Finalization of program requested. *)
   | "WrongCommunication"              (* Erroneous communication was received:
                                          this service message is used by the
                                          PIO library to warn the caller
                                          program.
                                          The [WrongCommunication] service
                                          implements a limited kind of exception
                                          raising, whether the implementation
                                          language supports exceptions or not. *)
*)

(** {6 Typed value description} *)

(** 
   For the time being, typed values are restricted to:
    - one boolean ("true" or "false"),
    - one string (a sequence of bytes),
    - one integer,
    - one float,
    - one boolean couple,
    - one boolean triple,
    - one integer couple,
    - one integer triple,
    - one float couple,
    - one float triple,
    - vectors of such,
    - float matrices.

   To prevent erroneous interpretation, a typed value contains its type.
   
   In addition, to enforce integrity of the communication, all typed values
   are delimited between two balanced markers.
*)

(**
   Typed_value ::=
     "begin" '\n'                      (* beginning of typed value marker *)
     "let" Ident_lexem " =" '\n'       (* optional name for the typed value *)
     Value_type '\n'                   (* the type of the following value *)
     Value                             (* a value of the preceding type *)
     "end" '\n' '\n'                   (* end of typed value marker *)
*)

(** {7 The type of values description} *)

(** The types of values is a simplified form of a real language type algebra,
    although types for [PIO] encapsulate some size information for arrays
    (i.e. integer values). *)

(**
   Value_type ::= Scalar_type | Array_type

   Array_type ::= Vector_type | Multi_d_type

   Vector_type ::= "[1" Scalar_type "1]"

   Multi_d_type ::= '[' Natural p Scalar_type Natural p ']'

   Scalar_type ::= Simple_type | Tuple_type

   Simple_type ::= Bool_type | String_type | Number_type

   Tuple_type ::=
   | '(' Simple_type ", " Simple_type ')'
                                       (* couple type *)
   | '(' Simple_type ", " Simple_type ", " Simple_type ')'
                                       (* triple type *)

   Bool_type ::=
   | "%B"  (* boolean type,
              declare this value as a C int on any architecture,
              this value is a Caml bool on any platform. *)

   String_type ::=
   | "%S"  (* string type,
              declare this value as a C char* on any architecture,
              this value is a Caml string on any platform. *)

   Number_type ::=
   | "%i"  (* integer type,
              declare this value as a C int on any architecture,
              this value is a Caml int on any platform. *)
   | "%ni" (* native integer with exactly one word of information:
              this value is a C pointer on any architecture, so
              declare this value as a C long on any architecture,
              (including LLP64 for which it could have been a long long),
              this value is a Caml Nativeint.t on any platform.
              Note: in case of communication from 32 bits to 64 bits
              architectures, the receiver must return the corresponding
              64 bits long; in the converse situation, if the integer
              can be truncated without loss of information, it must be
              truncated, otherwise the library must fail. *)
   | "%li" (* integer with exactly 32 bits of information,
              declare this value as a C int on any architecture,
              (as a consequence, on an ILP64 architecture the C
              programmer should not use type _int32),
              this value is a Caml Int32.t on any platform. *)
   | "%Li" (* integer with exactly 64 bits of information,
              a C long long on a 32 bits architecture,
              a C long on a 64 bits architecture (or a long long on LLP64),
              this value is a Caml Int64.t on any platform. *)
   | "%f"  (* float type in usual human readable representation,
              declare this value as a C double on any architecture,
              this value is a Caml float on any platform. *)
   | "%bf" (* float type in binary representation,
              declare this value as a C double on any architecture,
              this value is a Caml float on any platform. *)
*)

(** {7 Value description} *)

(** {8 Simple lexems} *)

(** Simple lexems are simple Caml lexems, either string or number lexems. *)

(**
   Simple_lexem ::=
   | Ident_lexem           (* A reference to an already named typed value *)
   | Bool_lexem | String_lexem | Number_lexem
*)

(** {9 Ident lexems} *)

(** An ident lexem is a simple form of identifier, which is also a Caml value
  ident. *)

(**
   Ident_lexem ::= [_a-z][0-9_a-z]*
     (* A Pio value ident. *) *)

(** {9 Boolean lexems} *)

(** A boolean lexem is just a Caml bool. *)

(**
   Bool_lexem ::= "true" | "false"
     (* A valid Caml boolean. *) *)

(** {9 Character string lexems} *)

(** A string lexem is just a Caml string incorporating its size. *)

(**
   String_lexem ::= Size '\n' String
     (* A delimited string with its size.
        Quoted characters are counted as only one character in the size. *)

   String ::= '"' Char* '"'
     (* A valid Caml delimited string. *)

   Char ::= Escaped_char | Printable_char
     (* A valid Caml character. *)

   Escaped_char ::=
   | '\\' Decimal_digit Decimal_digit Decimal_digit
     (* The decimal value of the ASCII code of the character
        (between 0 and 255). *)
   | "\\b" | "\\n" | "\\r" | "\\t"
     (* The backspace (BS, ASCII 8), linefeed (LF, ASCII 10),
        carriage return (CR, ASCII 13),
        and horizontal tabulation (TAB, ASCII 9) characters. *)
   | "\\\\" | "\\\"" | "\\\'" | "\\\ "
     (* The baskslash (ASCII 92), double quote (ASCII 34),
        single quote (ASCII 39) and space (SPC, ASCII 32) characters. *)

   Printable_char ::= '[32-33]' | '[35-91]' | '[93-126]'
     (* The set of printable characters (letters, digits, punctation
        characters, spaces). More precisely, printable characters are
        characters whose ASCII code ranges from 32 to 126, excluding 34 (double
        quote) and 92 (backslash).
        Note: printable characters can be escaped (i.e., considered as
        [Escaped_char] and encoded as a 3 decimal digits ASCII code). *)
*)

(** {9 Number lexems} *)

(** Roughly speaking numbers are Caml lexems for numbers. *)

(**
   Number_lexem ::= Integer | Float | Binary_number

   Integer ::= ``a valid Caml lexem: an integer''
   Float ::= ``a valid Caml lexem: a floating point number''
     (* See the Caml documentation for those specifications. *)

   Binary_number ::= '&' Size ``size bytes''
     (* The encoding of a floating-point number in binary representation.
        The Size is the number of bytes of the binary representation.
        (For instance, size is 8 for the 64 bits encoding representation of a
         floating-point number according to the IEEE 754 standard.) *)

*)

(** {8 Scalars} *)

(** A [Scalar] is either a simple lexem or a tuple with the usual syntax. *)

(**
   Scalar ::= Simple_lexem | Tuple_lexem

   Tuple_lexem ::=
   | '(' Simple_lexem ", " Simple_lexem ')'
   | '(' Simple_lexem ", " Simple_lexem ", " Simple_lexem ')'
*)

(** {8 Values} *)

(** A [Value] is either a scalar or an array. *)

(**
   Value ::= Scalar_item | Array_item

   Scalar_item ::= Scalar ';' '\n'
   Array_item ::=  Array ';' '\n'
*)

(** {8 Array values} *)

(**
   Array ::= Vector | Multi_d

   Vector ::=
     "[1" '\n'                (* beginning of vector marker *)
     Size '\n'                (* size of the vector *)
     Scalar_item*             (* items of the vector *)
     "1]"                     (* end of vector marker *)

   Multi_d ::=
     '[' Natural p '\n'       (* beginning of multi-d array marker,
                                 with [p] > 1 *)
     Multi_d_sizes  '\n'      (* the sizes of the multi-d array,
                                 i.e. the [p] dimensions of the multi-d array *)
     Multi_d_layout '\n'      (* the layout of the multi-d array *)
     Line_item*
       (* The 1D-lines of the multi-d array.
          When $\Pi_{i=1}^p dim_i = 0$, there is no line here.
          Otherwise, the number of lines is $\Pi_{i=2}^p dim_{s(i)}$ where
          $(dim_1, dim_2, ..., dim_p)$ is the size of the multi-d array and $s$
          is the permutation defining the layout *)
     Natural p ']'            (* end of multi-d array marker *)

   Multi_d_sizes ::= '<' Natural_list2 '>'

   Natural_list2 ::=          (* a comma-separated list of at least 2 integers *)
   | Natural ", " Natural     (* the two last integers *)
   | Natural ", " Natural_list2
                              (* an integer followed by the rest of the list *)

   Multi_d_layout ::=
   | Natural_list2
       (* the permutation of the range [0..p-1] expressing the layout *)
   | 'C'
       (* C-like or row-major layout,
          corresponding to the identity permutation *)
   | 'F'
       (* Fortran-like or column-major layout,
          corresponding to the reverse order permutation *)

   Line ::=
     "[|" '\n'                (* beginning of line marker *)
     Scalar_item*             (* items of the line *)
     "|]"                     (* end of line marker *)

   Line_item ::= Line ';' '\n'
*)

(** {8 An example: matrices specification} *)

(** Rectangular arrays, or matrices, are particular cases of multi dimensional
    arrays, more precisely 2D-arrays. The concrete syntax for matrices will
    thus be: *)

(**
   Matrix ::=
     "[2" '\n'                (* beginning of matrix marker *)
     '<' Natural ", " Natural '>' '\n'
                              (* size of the matrix, rows then columns *)
     Array_layout '\n'        (* layout of the matrix *)
     Line_item*               (* rows, or columns, of the matrix *)
     "2]"                     (* end of matrix marker *)
*)

(** {6 The Caml API for [PIO]} *)

(** We describe the communication library between Caml programs and foreign
    programs. *)

(** {7 Naming conventions} *)

(**
    The Caml [PIO] library provides input/output primitives operating on
    channels.
    The naming convention is:
    - primitives for reading a <type> from a given input channel are named
      [input_]<type>,
    - primitives for writing a <type> into a given output channel are named
      [output_]<type>.

    The Caml [PIO] library provides input/output primitives operating on
    buffers.
    The naming convention is:
    - primitives for reading a <type> from a given input buffer
      (a [Scanf.Scanning.scanbuf] value) are named [read_]<type>,
    - primitives for writing a <type> into a given output buffer
      (a [Buffer.t] value) are named [write_]<type>.

    In addition, the [PIO] library provides acknowledged input/output
    primitives operating on buffers. Acknowledged input/output primitives
    involve an acknowledgment mechanism to ensure that the communication has
    been properly delivered (indeed read or indeed written).
    The naming convention is:
    - primitives for acknowledged reading a <type> from a given input buffer
      (a [Scanf.Scanning.scanbuf] value) are named [receive_]<type>,
    - primitives for acknowledged writing a <type> into a given output buffer
      (a [Buffer.t] value) are named [send_]<type>.

    For the time being, the acknowledged level is not yet implemented and the
    [receive_] and [send_] primitives are equivalent to their [read_] and
    [write_] counterparts.
 *)

(** {7 Output communication functions} *)

(** {8 Functions to send communications} *)

(** Those functions send a communication to a buffer.

    More precisely, they write a {!Communication.t} value on a Buffer.t
    output buffer. *)

val send_one_value : Buffer.t -> Typed_value.t -> unit;;
(** [send_one_value ob v] sends to the Buffer.t output buffer [ob] a
  [Phrase] communication containing the unique {!Typed_value.t} value [v]. *)

val send_values : Buffer.t -> Typed_value.t array -> unit;;
(** [send_values ob vals] sends to the Buffer.t output buffer [ob] a
  [Phrase] communication containing all the typed values of the vector of
  {!Typed_value.t} values [vals]. *)

val send_communication : Buffer.t -> Communication.t -> unit;;
(** [send_communication ob com] sends its {!Communication.t} [comm] argument
  to the Buffer.t output buffer [ob]. *)

(** {8 Functions to send typed values} *)

(** Those functions send a regular Caml value to a buffer.

    More precisely, they write the {!Typed_value.t} value corresponding to
    their Caml value argument on a Buffer.t output buffer. *)

val send_string : Buffer.t -> string -> unit;;
(** [Io.send_string ob s] sends the string [s] as a {!Typed_value.t} typed
    value to the Buffer.t output buffer [ob]. *)

val send_int : Buffer.t -> int -> unit;;
(** Same as above, for an integer. *)

val send_int_couple : Buffer.t -> int * int -> unit;;
(** Same as above, for an integer couple. *)

val send_int_triple : Buffer.t -> int * int * int -> unit;;
(** Same as above, for an integer triple. *)

val send_float : Buffer.t -> float -> unit;;
(** Same as above, for a float. *)

val send_float_couple : Buffer.t -> float * float -> unit;;
(** Same as above, for a float couple. *)

val send_float_triple : Buffer.t -> float * float * float -> unit;;
(** Same as above, for a float triple. *)

val send_string_vector : Buffer.t -> string array -> unit;;
(** Same as above, for a string array. *)

val send_int_vector : Buffer.t -> int array -> unit;;
(** Same as above, for an integer array. *)

val send_int_couple_vector : Buffer.t -> (int * int) array -> unit;;
(** Same as above, for an integer couple array. *)

val send_int_triple_vector : Buffer.t -> (int * int * int) array -> unit;;
(** Same as above, for an integer triple array. *)

val send_float_vector : Buffer.t -> float array -> unit;;
(** Same as above, for a float array. *)

val send_float_couple_vector : Buffer.t -> (float * float) array -> unit;;
(** Same as above, for a float couple array. *)

val send_float_triple_vector :
    Buffer.t -> (float * float * float) array -> unit
;;
(** Same as above, for a float triple array. *)

val send_c_float_matrix : Buffer.t -> float Umatrix.t -> unit;;
(** Same as above, for a C float matrix. *)

val send_fortran_float_matrix : Buffer.t -> float Umatrix.t -> unit;;
(** Same as above, for a Fortran float matrix. *)

(** {7 Input communication functions} *)

exception Wrong_communication of string * string;;

(** {8 Functions to receive communications} *)

(** Those functions receive a communication from an input buffer,
  and return the corresponding Caml value. *)

val receive_communication : Scanf.Scanning.scanbuf -> Communication.t;;
(** [receive_communication ib] reads a communication from the
  [!Scanf.Scanning.scanbuf] scanning buffer [ib]. *)

val receive_one_value : Scanf.Scanning.scanbuf -> Typed_value.t;;
(** [receive_one_value ib] reads a [!Communication.t] communication from the
  [!Scanf.Scanning.scanbuf] input buffer [ib]. The communication should be a
  [Phrase] with only one typed value, and the function returns this
  [!Typed_value.t] value. *)

val receive_values : Scanf.Scanning.scanbuf -> Typed_value.t array;;
(** [receive_values ib] reads a [Phrase] communication from the
  [!Scanf.Scanning.scanbuf] input buffer [ib] and returns the corresponding
  vector of typed values. *)

(** {8 Functions to receive typed values} *)

(** Those functions receives a regular Caml value from an input buffer.

    More precisely, they return the regular Caml value corresponding to the
    {!Typed_value.t} value they read from their {Scanf.Scanning.scanbuf} input
    buffer argument. *)

val receive_string : Scanf.Scanning.scanbuf -> string;;
(** [Io.receive_string ib] receives a typed value from the input buffer [ib]
    and converts it to a regular Caml [string] value. *)

val receive_int : Scanf.Scanning.scanbuf -> int;;
(** Same as above, for an integer. *)

val receive_int_couple : Scanf.Scanning.scanbuf -> int * int;;
(** Same as above, for an integer couple. *)

val receive_int_triple : Scanf.Scanning.scanbuf -> int * int * int;;
(** Same as above, for an integer triple. *)

val receive_float : Scanf.Scanning.scanbuf -> float;;
(** Same as above, for a float. *)

val receive_float_couple : Scanf.Scanning.scanbuf -> float * float;;
(** Same as above, for a float couple. *)

val receive_float_triple : Scanf.Scanning.scanbuf -> float * float * float;;
(** Same as above, for a float triple. *)

val receive_string_vector : Scanf.Scanning.scanbuf -> string array;;
(** Same as above, for a string array. *)

val receive_int_vector : Scanf.Scanning.scanbuf -> int array;;
(** Same as above, for an integer array. *)

val receive_int_couple_vector : Scanf.Scanning.scanbuf -> (int * int) array;;
(** Same as above, for an integer couple array. *)

val receive_int_triple_vector :
    Scanf.Scanning.scanbuf -> (int *  int * int) array
;;
(** Same as above, for an integer triple array. *)

val receive_float_vector : Scanf.Scanning.scanbuf -> float array;;
(** Same as above, for a float array. *)

val receive_float_couple_vector :
    Scanf.Scanning.scanbuf -> (float * float) array
;;
(** Same as above, for a float couple array. *)

val receive_float_triple_vector :
    Scanf.Scanning.scanbuf -> (float * float * float) array
;;
(** Same as above, for a float triple array. *)

val receive_c_float_matrix : Scanf.Scanning.scanbuf -> float Umatrix.t;;
(** Same as above, for a C float matrix. *)

val receive_fortran_float_matrix : Scanf.Scanning.scanbuf -> float Umatrix.t;;
(** Same as above, for a Fortran float matrix. *)
