(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                  projet Estime, INRIA-Rocquencourt                  *)
(*                                                                     *)
(*  Copyright 2007-2011 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: batch.ml,v 1.10.2.1 2011-09-21 13:08:04 fclement Exp $ *)

(* Module used to spawn batch tasks *)
module type FIFO = sig
  type 'a t;;

  val empty : 'a t;;
  val unshift : 'a t -> 'a -> 'a t;;
  val shift : 'a t -> ('a option * 'a t)
end
;;

module Fifo : FIFO = struct
  type 'a t = {
    inp : 'a list;
    outp : 'a list;
  };;

  let empty = { inp = []; outp = [] };;

  let unshift t x =
    { t with inp = x :: t.inp };;

  let shift t =
    let t =
      if t.outp = []
      then { outp = List.rev t.inp; inp = []; }
      else t in
    match t.outp with
    | x :: xs -> (Some x, { t with outp = xs })
    | [] -> (None, t)
  ;;
end
;;

type shell_action = {
  program : string;
  args : string array;
};;

type fail_action =
  | Fail_error of string
  | Fail_warn of string
  | Fail_ignore
;;

type task_type =
  | Background
  | Normal
  | Timeouted of int
;;

type pid = int;;

type task = {
  cmd : command; (* Command to launch to run the task *)
  on_fail : fail_action; (* What to do if it fails *)
}
and command =
  | Batch of batch
  | Control of (pid list -> batch)
  | Echo of string
  | Fail
  | Pipe of (batch * (in_channel -> batch))
  | Shell of (shell_action * task_type)
and batch = task Fifo.t
;;

let empty =
  Fifo.empty
;;

let add_shell_task batch fail cmd timeout =
  Fifo.unshift batch { cmd = Shell (cmd, timeout); on_fail = fail; }
;;

let add_subscript_task batch fail subbatch =
  Fifo.unshift batch { cmd = Batch subbatch; on_fail = fail; }
;;

let add_control_task batch fail c =
  Fifo.unshift batch { cmd = Control c; on_fail = fail; }
;;

let add_pipe_task batch fail subbatch f =
  Fifo.unshift batch { cmd = Pipe (subbatch, f); on_fail = fail; }
;;

let add_echo_task batch msg =
  Fifo.unshift batch { cmd = Echo msg; on_fail = Fail_ignore; }
;;

let add_fail_task batch msg =
  Fifo.unshift batch { cmd = Fail; on_fail = Fail_error msg; }
;;

type run_return =
  | Err of int
  | Success
  | Timeout
;;

type batch_error =
  | Batch_timeout of shell_action
  | Batch_shell_error of (int * shell_action)
  | Batch_failure
;;

type batch_run_return =
  | Ok of pid list
  | Error of (pid list * batch_error)
;;

let execvp_io p a oc =
  List.iter (Unix.dup2 oc)
    [ Unix.stdout; (* Unix.stderr; *) ];
  try Unix.execvp p a
  with Unix.Unix_error (e, _, _) ->
    let cmd = String.concat " " (Array.to_list a) in
    Printf.eprintf "Unix error during execvp(%S): %s\n"
      cmd (Unix.error_message e);
    exit 2
;;

let run_timeouted_command timeout oc cmd =
  let cpid = Unix.fork () in
  if cpid = 0 then
    let spawn { program = p; args = a; } =
      let cpid = Unix.fork () in
      let kill _ = Unix.kill cpid Sys.sigkill in
      if cpid = 0 then execvp_io p a oc else begin
        Sys.set_signal Sys.sigalrm (Sys.Signal_handle kill);
        ignore (Unix.alarm timeout);
        let (_, status) =
          let rec loop () =
            try Unix.waitpid [] cpid
            with Unix.Unix_error (Unix.EINTR ,_, _) -> loop () in
          loop () in
        Sys.set_signal Sys.sigalrm Sys.Signal_default;
        let ret_code =
          match status with
          | Unix.WEXITED i -> i
          | Unix.WSIGNALED _ -> 255
          | Unix.WSTOPPED _ -> kill 0; 255 in
        exit ret_code
      end in
    spawn cmd
  else begin
    let (_, status) = Unix.waitpid [] cpid in
    match status with
    | Unix.WEXITED ret ->
      if ret = 0 then Ok [] else
      if ret = 255 then Error ([], Batch_timeout cmd) else
      Error ([], Batch_shell_error (ret, cmd))
    | Unix.WSIGNALED _ | Unix.WSTOPPED _ ->
      Error ([], Batch_shell_error (0, cmd))
  end
;;

let run_background_command oc cmd =
  let cpid = Unix.fork () in
  if cpid = 0
  then execvp_io cmd.program cmd.args oc
  else Ok [cpid]
;;

let run_normal_command oc cmd =
  let cpid = Unix.fork () in
  if cpid = 0
  then execvp_io cmd.program cmd.args oc
  else begin
    let (_, status) = Unix.waitpid [] cpid in
    match status with
    | Unix.WEXITED ret ->
      if ret = 0 then Ok [] else
      Error ([], Batch_shell_error (ret, cmd))
    | Unix.WSIGNALED _ | Unix.WSTOPPED _ ->
      Error ([], Batch_shell_error (0, cmd))
  end
;;

let run_command t oc =
  let oc = Unix.descr_of_out_channel oc in
  match t with
  | Background -> run_background_command oc
  | Normal -> run_normal_command oc
  | Timeouted t -> run_timeouted_command t oc
;;

let rec eval_batch b oc ec =
  let mk_pipe () =
    let (ic, oc) = Unix.pipe () in
    (Unix.in_channel_of_descr ic, Unix.out_channel_of_descr oc) in
  let eval_pipe b f =
    let (ic, oc_pipe) = mk_pipe () in
    match eval_batch b oc_pipe ec with
    | Ok l ->
      begin match eval_batch (f ic) oc ec with
      | Ok l' -> Ok (l @ l')
      | Error (l', e) -> Error (l @ l', e)
      end
    | Error _ as err -> err in
  let eval_command cmd pids =
    match cmd with
    | Shell (cmd, t) -> run_command t oc cmd
    | Control f -> eval_batch (f pids) oc ec
    | Echo m -> Printf.fprintf ec "%s\n" m; Ok []
    | Fail -> Error ([], Batch_failure)
    | Pipe (b, f) -> eval_pipe b f
    | Batch b -> eval_batch b oc ec in
  let rec loop b pids =
    match Fifo.shift b with
    | (Some { cmd = cmd; on_fail = f; }, b) ->
      begin match eval_command cmd pids with
      | Ok l -> loop b (l @ pids)
      | Error (l, e) ->
        begin match f with
        | Fail_error em -> Printf.fprintf ec "%s\n" em; Error (l @ pids, e)
        | Fail_warn em -> Printf.fprintf ec "%s\n" em; loop b pids
        | Fail_ignore -> loop b pids
        end
      end
    | (None, _) -> Ok pids in
  loop b []
;;
